@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2004 Mads Martin Joergensen <mmj at mmj.dk>
- * Copyright (C) 2023 Baptiste Daroussin <bapt@FreeBSD.org>
+ * Copyright (C) 2023-2025 Baptiste Daroussin <bapt@FreeBSD.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -25,8 +25,8 @@
 
 #include "mlmmj.h" /* For struct mailhdr and struct strlist */
 
-bool findit(const char *line, const strlist *headers);
-void getinfo(const char *line, struct mailhdr *readhdrs);
 int do_all_the_voodoo_here(int infd, int outfd, int hdrfd, int footfd,
 	      const strlist *delhdrs, struct mailhdr *readhdrs,
 	      strlist *allhdrs, const char *subjectprefix, int replyto);
+void scan_headers(FILE *f, struct mailhdr *readhdrs, strlist *allhdrs,
+	strlist *allunfoldeds);
