@@ -1,6 +1,5 @@
 /* Copyright (C) 2004 Mads Martin Joergensen <mmj at mmj.dk>
- *
- * $Id$
+ * Copyright (C) 2025 Baptiste Daroussin <bapt@FreeBSD.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -34,37 +33,43 @@
 #include "do_all_the_voodoo_here.h"
 #include "log_error.h"
 #include "wrappers.h"
+#include "find_email_adr.h"
+#include "utils.h"
 
-bool
-findit(const char *line, const strlist *headers)
+
+/* read line and populate a struct to save a list important header see mlmmj-process.c:254 and mlmmj.h:97 */
+
+void
+scan_headers(FILE *f, struct mailhdr *readhdrs, strlist *allhdrs, strlist *allunfoldeds)
 {
-	size_t len;
+	char *hdrline, *unfolded;
 
-	tll_foreach(*headers, it) {
-		len = strlen(it->item);
-		if(strncasecmp(line, it->item, len) == 0)
-			return true;
-	}
+	/* scan all headers and populate readhdrs , allhdrs , allunfoldeds */
+	for (;;) {
+	        /* hdrline contains the header in one line and unfolded contains the original data */
+		hdrline = gethdrline(f, &unfolded);
 
-	return false;
-}
-
-void getinfo(const char *line, struct mailhdr *readhdrs)
-{
-	int i = 0;
-	size_t tokenlen;
-
-	while(readhdrs[i].token) {
-		tokenlen = strlen(readhdrs[i].token);
-		if(strncasecmp(line, readhdrs[i].token, tokenlen) == 0) {
-			readhdrs[i].valuecount++;
-			readhdrs[i].values =
-				(char **)xrealloc(readhdrs[i].values,
-				  readhdrs[i].valuecount * sizeof(char *));
-			readhdrs[i].values[readhdrs[i].valuecount - 1] =
-				xstrdup(line + tokenlen);
+		/* end of headers */ 
+		if(hdrline == NULL) {
+			break;
 		}
-		i++;
+
+		/* Do we want info from hdrs? Get it before it's gone */
+		if(readhdrs)
+			save_hdr_if_needed(hdrline, readhdrs);
+
+		if (allhdrs != NULL) {
+			/* Snatch a copy of the header */
+			tll_push_back(*allhdrs, xstrdup(hdrline));
+		}
+		
+		if (allunfoldeds != NULL) {
+			/* Snatch a copy of unfolded */
+			tll_push_back(*allunfoldeds, xstrdup(unfolded));
+		}
+		
+		free(hdrline);
+		free(unfolded);
 	}
 }
 
@@ -73,23 +78,38 @@ int do_all_the_voodoo_here(int infd, int outfd, int hdrfd, int footfd,
 		 strlist *allhdrs, const char *prefix, int replyto)
 {
 	char *hdrline, *unfolded, *unqp;
+	strlist allunfoldeds = tll_init();
+	char *posteraddr = NULL;
 	bool hdrsadded = false;
 	bool subject_present = false;
 	FILE *f = fdopen(dup(infd), "r");
 
-	for(;;) {
-		hdrline = gethdrline(f, &unfolded);
+	scan_headers(f, readhdrs, allhdrs, &allunfoldeds);
+
+	/* if we have custom headers , extract posteraddr from current header */
+	if(hdrfd >= 0) {
+		strlist fromemails = tll_init();
+		if ( readhdrs[0].valuecount == 1 ) {
+		        find_email_adr(readhdrs[0].values[0], &fromemails);
+			posteraddr = tll_front(fromemails);
+		}
+	}
+	/* scan all headers from allhdrs , and do the real things */
+	tll_foreach(*allhdrs, a_header ) {
+
+	        /* hdrline contains the unfolded value of the header and unfolded contains the original data */
+	        hdrline=a_header->item ;
+		unfolded=tll_pop_front(allunfoldeds) ;
 
 		/* add extra headers before MIME* headers,
 		   or after all headers */
 		if(!hdrsadded &&
-		    (hdrline == NULL || strncasecmp(hdrline, "mime", 4) == 0)) {
+		    ( strncasecmp(hdrline, "mime", 4) == 0)) {
 			if(hdrfd >= 0) {
-				if(dumpfd2fd(hdrfd, outfd) < 0) {
+			        if(process_headers_fd(hdrfd,outfd,posteraddr) < 0) {
 					log_error(LOG_ARGS, "Could not "
 						"add extra headers");
-					free(hdrline);
-					free(unfolded);
+					tll_free_and_free(allunfoldeds, free);
 					return -1;
 				}
 				fsync(outfd);
@@ -97,33 +117,6 @@ int do_all_the_voodoo_here(int infd, int outfd, int hdrfd, int footfd,
 			hdrsadded = true;
 		}
 
-		/* end of headers */ 
-		if(hdrline == NULL) {
-			/* add Subject if none is present
-			   and a prefix is defined */
-			if (prefix && !subject_present) {
-				dprintf(outfd, "Subject: %s\n", prefix);
-				subject_present = true;
-			}
-			/* write LF */
-			if(dprintf(outfd, "\n") < 0) {
-				free(hdrline);
-				free(unfolded);
-				log_error(LOG_ARGS, "Error writing hdrs.");
-				return -1;
-			}
-			free(hdrline);
-			free(unfolded);
-			break;
-		}
-
-		/* Do we want info from hdrs? Get it before it's gone */
-		if(readhdrs)
-			getinfo(hdrline, readhdrs);
-
-		/* Snatch a copy of the header */
-		tll_push_back(*allhdrs, xstrdup(hdrline));
-
 		/* Add Subject: prefix if wanted */
 		if(prefix) {
 			if(strncasecmp(hdrline, "Subject:", 8) == 0) {
@@ -132,7 +125,6 @@ int do_all_the_voodoo_here(int infd, int outfd, int hdrfd, int footfd,
 				if(strstr(hdrline + 8, prefix) == NULL &&
 				   strstr(unqp, prefix) == NULL) {
 					dprintf(outfd, "Subject: %s%s\n", prefix, hdrline + 8);
-					free(hdrline);
 					free(unqp);
 					continue;
 				}
@@ -141,7 +133,7 @@ int do_all_the_voodoo_here(int infd, int outfd, int hdrfd, int footfd,
 		}
 
 		/* Should it be stripped? */
-		if(!delhdrs || !findit(hdrline, delhdrs))
+		if(!delhdrs || !line_startwithany(hdrline, delhdrs, true))
 			dprintf(outfd, "%s", unfolded);
 
 		/* Should Reply-To be added? */
@@ -150,11 +142,33 @@ int do_all_the_voodoo_here(int infd, int outfd, int hdrfd, int footfd,
 				dprintf(outfd, "Reply-To:%s\n", hdrline + 5);
 			}
 		}
-
-		free(hdrline);
-		free(unfolded);
 	}
 
+	if(!hdrsadded ) {
+		if(hdrfd >= 0) {
+		  if(process_headers_fd(hdrfd,outfd,posteraddr) < 0) {
+				log_error(LOG_ARGS, "Could not "
+					"add extra headers");
+				tll_free_and_free(allunfoldeds, free);
+				return -1;
+			}
+			fsync(outfd);
+		}
+	}
+
+	if (prefix && !subject_present) {
+		dprintf(outfd, "Subject: %s\n", prefix);
+		subject_present = true;
+	}
+	/* write LF */
+	if(dprintf(outfd, "\n") < 0) {
+	        tll_free_and_free(allunfoldeds, free);
+		log_error(LOG_ARGS, "Error writing hdrs.");
+		return -1;
+	}
+	
+	tll_free_and_free(allunfoldeds, free);
+
 	lseek(infd, ftello(f), SEEK_SET);
 	/* Just print the rest of the mail */
 	if(dumpfd2fd(infd, outfd) < 0) {
