@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2003 Mads Martin Joergensen <mmj at mmj.dk>
- * Copyright (C) 2023 Baptiste Daroussin <bapt@FreeBSD.org>
+ * Copyright (C) 2023-2025 Baptiste Daroussin <bapt@FreeBSD.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -110,7 +110,7 @@ is_valid_email(const char *email, const char *log)
 	return (true);
 }
 
-static void
+__dead2 static void
 lc_sub(struct listcontrol *lc, struct ml *ml, struct sub *sub)
 {
 	if (!is_valid_email(lc->fromemail, sub->typereq))
@@ -219,7 +219,7 @@ int listcontrol(strlist *fromemails, struct ml *ml, const char *controlstr,
 	if (cc == NULL)
 		return (-1);
 
-	if(tll_length(*fromemails) != 1 && cc->type != CTRL_BOUNCES) {
+	if (tll_length(*fromemails) != 1 && cc->type != CTRL_BOUNCES) {
 		errno = 0;
 		log_error(LOG_ARGS, "Ignoring mail with invalid From: "
 			"which was not a bounce: %d", tll_length(*fromemails));
@@ -261,7 +261,7 @@ int listcontrol(strlist *fromemails, struct ml *ml, const char *controlstr,
 	case CTRL_SUBSCRIBE_NOMAIL:
 		if (sub.typereq == NULL) {
 			sub.typereq = "A subscribe-nomail";
-			sub.ctrlstr = "nomailsub";
+			sub.ctrlstr = "nonomailsub";
 			sub.typesub = SUB_NOMAIL;
 			sub.type = "nomail";
 			sub.typedeny = "sub-deny-nomail";
@@ -394,12 +394,10 @@ int listcontrol(strlist *fromemails, struct ml *ml, const char *controlstr,
 		bret = bouncemail(ml->fd, lowercase(c), bouncenr);
 		if (bret == BOUNCE_DONE)
 			save_lastbouncedmsg(ml->fd, c, mailname);
-		if (bouncemail(ml->fd, c, bouncenr) == BOUNCE_FAIL)
+		if (bret == BOUNCE_FAIL)
 			exit(EXIT_FAILURE);
-		if (bret == BOUNCE_OK)
-			unlink(mailname);
 		unlink(mailname);
-		exit(bret == BOUNCE_FAIL ? EXIT_FAILURE : EXIT_SUCCESS);
+		exit(EXIT_SUCCESS);
 		break;
 
 	/* listname+release-COOKIE@domain.tld */
@@ -408,7 +406,7 @@ int listcontrol(strlist *fromemails, struct ml *ml, const char *controlstr,
 	/* DEPRECATED: listname+moderate-subscribeCOOKIE@domain.tld */
 	case CTRL_MODERATE:
 		/* Subscriber moderation; DEPRECATED */
-		if(strncmp(param, "subscribe", 9) == 0) {
+		if (strncmp(param, "subscribe", 9) == 0) {
 			tmpstr = xstrdup(param + 9);
 			free(param);
 			param = tmpstr;
@@ -418,7 +416,7 @@ int listcontrol(strlist *fromemails, struct ml *ml, const char *controlstr,
 		xasprintf(&moderatefilename, "moderation/%s", param);
 		xasprintf(&sendfilename, "%s.sending", moderatefilename);
 		/* Rename it to avoid mail being sent twice */
-		if(renameat(ml->fd, moderatefilename, ml->fd, sendfilename) != 0) {
+		if (renameat(ml->fd, moderatefilename, ml->fd, sendfilename) != 0) {
 			if (errno == ENOENT) {
 				errno = 0;
 				log_error(LOG_ARGS, "A release request was"
@@ -439,6 +437,15 @@ int listcontrol(strlist *fromemails, struct ml *ml, const char *controlstr,
 			unlinkat(ml->fd, omitfilename, 0);
 		free(omitfilename);
 		free(moderatefilename);
+		bool autosubscribe = statctrl(ml->ctrlfd, "autosubscribe");
+		if (autosubscribe) {
+			int mfd = openat(ml->fd, sendfilename, O_RDONLY);
+			if (mfd == -1) {
+				free(sendfilename);
+				return (-1);
+			}
+			autosubscribe_sender(ml, mfd, SUB_NOMAIL);
+		}
 
 		log_oper(ml->fd, OPLOGFNAME, "%s released %s", lc.fromemail, param);
 
@@ -560,7 +567,7 @@ permit:
 			return -1;
 		}
 		if (statctrl(ml->ctrlfd, "subonlyget")) {
-			if(is_subbed(ml->fd, tll_front(*fromemails), 0) ==
+			if (is_subbed(ml->fd, tll_front(*fromemails), 0) ==
 					SUB_NONE) {
 				errno = 0;
 				log_error(LOG_ARGS, "A get request was sent"
@@ -599,7 +606,7 @@ permit:
 
 	/* listname+list@domain.tld */
 	case CTRL_LIST:
-		if(statctrl(ml->ctrlfd, "nolistsubsemail"))
+		if (statctrl(ml->ctrlfd, "nolistsubsemail"))
 			return -1;
 		const char *owner = tll_front(*fromemails);
 		if (!ctrlvalues_contains(ml->ctrlfd, "owner", owner, false)) {
