@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2003, 2003, 2004 Mads Martin Joergensen <mmj at mmj.dk>
- * Copyright (C) 2023 Baptiste Daroussin <bapt@FreeBSD.org>
+ * Copyright (C) 2023-2025 Baptiste Daroussin <bapt@FreeBSD.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -56,29 +56,12 @@
 #include "send_help.h"
 #include "send_mail.h"
 
-enum action {
-	ALLOW,
-	SEND,
-	DENY,
-	MODERATE,
-	DISCARD
-};
-
-
-static char *action_strs[] = {
-	"allowed",
-	"sent",
-	"denied",
-	"moderated",
-	"discarded"
-};
-
-
 enum modreason {
 	MODNONSUBPOSTS,
 	MODNONMODPOSTS,
 	ACCESS,
-	MODERATED
+	MODERATED,
+	SELFMODERATED
 };
 
 
@@ -86,9 +69,24 @@ static char *modreason_strs[] = {
 	"modnonsubposts",
 	"modnonmodposts",
 	"access",
-	"moderated"
+	"moderated",
+	"selfmoderated"
 };
 
+static void
+free_parsed_hdrs(struct mailhdr *readhdrs, strlist fromemails,
+    strlist originalfromemails, strlist toemails, strlist ccemails,
+    strlist rpemails, strlist dtemails, strlist allheaders)
+{
+	free_mailhdrs(readhdrs);
+	tll_free_and_free(fromemails, free);
+	tll_free_and_free(originalfromemails, free);
+	tll_free_and_free(toemails, free);
+	tll_free_and_free(ccemails, free);
+	tll_free_and_free(rpemails, free);
+	tll_free_and_free(dtemails, free);
+	tll_free_and_free(allheaders, free);
+}
 
 static bool is_moderator(int listfd, const char *address,
 		char **moderators) {
@@ -98,7 +96,7 @@ static bool is_moderator(int listfd, const char *address,
 	xstring *str = NULL;
 	FILE *fp;
 
-	if((moderatorsfd = openat(listfd, "control/moderators", O_RDONLY)) < 0) {
+	if ((moderatorsfd = openat(listfd, "control/moderators", O_RDONLY)) < 0) {
 		log_error(LOG_ARGS, "Could not open 'control/moderators'");
 		exit(EXIT_FAILURE);
 	}
@@ -106,7 +104,7 @@ static bool is_moderator(int listfd, const char *address,
 
 	while (getline(&line, &linecap, fp) > 0) {
 		chomp(line);
-		if(address && strcasecmp(line, address) == 0) {
+		if (address && strcasecmp(line, address) == 0) {
 			foundaddr = true;
 			if (moderators == NULL)
 				goto out;
@@ -140,19 +138,20 @@ static void newmoderated(struct ml *ml, const char *mailfilename,
 	const char *efromismod = NULL;
 	const char *mailbasename = mybasename(mailfilename);
 	int notifymod = 0;
+	bool selfmoderate = 0;
 	struct mail mail;
 #if 0
 	printf("mailfilename = [%s], mailbasename = [%s]\n", mailfilename,
 			                                     mailbasename);
 #endif
 
-	if(statctrl(ml->ctrlfd, "ifmodsendonlymodmoderate"))
+	if (statctrl(ml->ctrlfd, "ifmodsendonlymodmoderate"))
 		efromismod = efromsender;
 
-	if(!is_moderator(ml->fd, efromismod, &moderators))
+	if (!is_moderator(ml->fd, efromismod, &moderators))
 		efromismod = NULL;
 
-	if(efromismod) mls = init_memory_lines(efromismod);
+	if (efromismod) mls = init_memory_lines(efromismod);
 	else mls = init_memory_lines(moderators);
 
 	free(moderators);
@@ -161,7 +160,15 @@ static void newmoderated(struct ml *ml, const char *mailfilename,
 	gen_addr_cookie(reject, ml, "reject-", mailbasename);
 
 	gen_addr(from, ml, "owner");
-	xasprintf(&to, "%s-moderators@%s", ml->name, ml->fqdn);
+
+	selfmoderate = statctrl(ml->ctrlfd, "selfmoderate");
+
+	if (selfmoderate) {
+		to = xstrdup(posteraddr);
+		modreason = SELFMODERATED;
+	} else {
+		xasprintf(&to, "%s-moderators@%s", ml->name, ml->fqdn);
+	}
 
 	txt = open_text(ml->fd, "moderate", "post",
 			modreason_strs[modreason], NULL, "moderation");
@@ -169,10 +176,8 @@ static void newmoderated(struct ml *ml, const char *mailfilename,
 	register_default_unformatted(txt, ml);
 	register_unformatted(txt, "subject", subject);
 	register_unformatted(txt, "posteraddr", posteraddr);
-	register_unformatted(txt, "moderateaddr", replyto); /* DEPRECATED */
 	register_unformatted(txt, "releaseaddr", replyto);
 	register_unformatted(txt, "rejectaddr", reject);
-	register_unformatted(txt, "moderators", "%moderators%"); /* DEPRECATED */
 	register_formatted(txt, "moderators",
 			rewind_memory_lines, get_memory_line, mls);
 	register_originalmail(txt, mailfilename);
@@ -194,7 +199,6 @@ static void newmoderated(struct ml *ml, const char *mailfilename,
 		register_default_unformatted(txt, ml);
 		register_unformatted(txt, "subject", subject);
 		register_unformatted(txt, "posteraddr", posteraddr);
-		register_unformatted(txt, "moderators", "%moderators%"); /* DEPRECATED */
 		register_formatted(txt, "moderators",
 				rewind_memory_lines, get_memory_line, mls);
 		register_originalmail(txt, mailfilename);
@@ -223,140 +227,23 @@ static void newmoderated(struct ml *ml, const char *mailfilename,
 		fclose(mail.fp);
 		exit(EXIT_SUCCESS);
 	}
+
+	if (selfmoderate) {
+		memset(&mail, 0, sizeof(mail));
+		mail.from = from;
+		mail.to = to;
+		mail.fp = fopen(queuefilename, "r");
+		if (send_single_mail(&mail, ml, false))
+			unlink(queuefilename);
+		fclose(mail.fp);
+		exit(EXIT_SUCCESS);
+	}
+
 	exec_or_die(mlmmjsend, "-l", "2", "-L", ml->dir, "-F", from,
 	    "-m", queuefilename, NULL);
 }
 
 
-static enum action do_access(strlist *rule_strs, strlist *hdrs,
-		const char *from, int listfd)
-{
-	unsigned int match;
-	char *rule_ptr;
-	char errbuf[128];
-	int err;
-	enum action act;
-	unsigned int not;
-	regex_t regexp;
-	char *hdr;
-	size_t i = 0;
-
-	tll_foreach(*rule_strs, it) {
-		rule_ptr = it->item;
-		if (strncmp(rule_ptr, "allow", 5) == 0) {
-			rule_ptr += 5;
-			act = ALLOW;
-		} else if (strncmp(rule_ptr, "send", 4) == 0) {
-			rule_ptr += 4;
-			act = SEND;
-		} else if (strncmp(rule_ptr, "deny", 4) == 0) {
-			rule_ptr += 4;
-			act = DENY;
-		} else if (strncmp(rule_ptr, "moderate", 8) == 0) {
-			rule_ptr += 8;
-			act = MODERATE;
-		} else if (strncmp(rule_ptr, "discard", 7) == 0) {
-			rule_ptr += 7;
-			act = DISCARD;
-		} else {
-			errno = 0;
-			log_error(LOG_ARGS, "Unable to parse rule #%d \"%s\":"
-					" Missing action keyword. Denying post from \"%s\"",
-					i, it->item, from);
-			log_oper(listfd, OPLOGFNAME, "Unable to parse rule #%d \"%s\":"
-					" Missing action keyword. Denying post from \"%s\"",
-					i, it->item, from);
-			return DENY;
-		}
-
-		if (*rule_ptr == ' ') {
-			rule_ptr++;
-		} else if (*rule_ptr == '\0') {
-			/* the rule is a keyword and no regexp */
-			log_oper(listfd, OPLOGFNAME, "mlmmj-process: access -"
-					" A mail from \"%s\" was %s by rule #%d \"%s\"",
-					from, action_strs[act], i, it->item);
-			return act;
-		} else {
-			/* we must have space or end of string */
-			errno = 0;
-			log_error(LOG_ARGS, "Unable to parse rule #%d \"%s\":"
-					" Invalid character after action keyword."
-					" Denying post from \"%s\"", i, it->item, from);
-			log_oper(listfd, OPLOGFNAME, "Unable to parse rule #%d \"%s\":"
-					" Invalid character after action keyword."
-					" Denying post from \"%s\"", i, it->item, from);
-			return DENY;
-		}
-
-		if (*rule_ptr == '!') {
-			rule_ptr++;
-			not = 1;
-		} else {
-			not = 0;
-		}
-
-		/* remove unanchored ".*" from beginning of regexp to stop the
-		 * regexp matching to loop so long time it seems like it's
-		 * hanging */
-		if (strncmp(rule_ptr, "^.*", 3) == 0) {
-			rule_ptr += 3;
-		}
-		while (strncmp(rule_ptr, ".*", 2) == 0) {
-			rule_ptr += 2;
-		}
-
-		if ((err = regcomp(&regexp, rule_ptr,
-				REG_EXTENDED | REG_NOSUB | REG_ICASE))) {
-			regerror(err, &regexp, errbuf, sizeof(errbuf));
-			regfree(&regexp);
-			errno = 0;
-			log_error(LOG_ARGS, "regcomp() failed for rule #%d \"%s\""
-					" (message: '%s') (expression: '%s')"
-					" Denying post from \"%s\"",
-					i, it->item, errbuf, rule_ptr, from);
-			log_oper(listfd, OPLOGFNAME, "regcomp() failed for rule"
-					" #%d \"%s\" (message: '%s') (expression: '%s')"
-					" Denying post from \"%s\"",
-					i, it->item, errbuf, rule_ptr, from);
-			return DENY;
-		}
-
-		match = 0;
-		tll_foreach(*hdrs, header) {
-			if (regexec(&regexp, header->item, 0, NULL, 0)
-					== 0) {
-				match = 1;
-				hdr = header->item;
-				break;
-			}
-		}
-
-		regfree(&regexp);
-
-		if (match != not) {
-			if (match) {
-				log_oper(listfd, OPLOGFNAME, "mlmmj-process: access -"
-						" A mail from \"%s\" with header \"%s\" was %s by"
-						" rule #%d \"%s\"", from, hdr, action_strs[act],
-						i, it->item);
-			} else {
-				log_oper(listfd, OPLOGFNAME, "mlmmj-process: access -"
-						" A mail from \"%s\" was %s by rule #%d \"%s\""
-						" because no header matched.", from,
-						action_strs[act], i, it->item);
-			}
-			return act;
-		}
-		i++;
-	}
-
-	log_oper(listfd, OPLOGFNAME, "mlmmj-process: access -"
-			" A mail from \"%s\" didn't match any rules, and"
-			" was denied by default.", from);
-	return DENY;
-}
-
 static void print_help(const char *prg)
 {
 	printf("Usage: %s -L /path/to/list\n"
@@ -375,7 +262,7 @@ int main(int argc, char **argv)
 	int i, opt, moderated = 0, send = 0;
 	enum modreason modreason;
 	int hdrfd, footfd, rawmailfd, donemailfd, omitfd;
-	bool addr_in_to_or_cc, notmetoo;
+	bool addr_in_to_or_cc, notmetoo, replyto;
 	bool findaddress = false, intocc = false, noprocess = false;
 	int maxmailsize = 0;
 	bool subonlypost, modonlypost, modnonsubposts, foundaddr = false;
@@ -442,7 +329,7 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if(ml.dir == NULL || mailfile == NULL) {
+	if (ml.dir == NULL || mailfile == NULL) {
 		errx(EXIT_FAILURE, "You have to specify -L and -m\n"
 		    "%s -h for help", argv[0]);
 	}
@@ -461,7 +348,7 @@ int main(int argc, char **argv)
 
         } while ((donemailfd < 0) && (errno == EEXIST));
 
-	if(donemailfd < 0) {
+	if (donemailfd < 0) {
 		log_error(LOG_ARGS, "could not create %s", donemailname);
 		free(donemailname);
 		exit(EXIT_FAILURE);
@@ -469,7 +356,7 @@ int main(int argc, char **argv)
 #if 0
 	log_error(LOG_ARGS, "donemailname = [%s]\n", donemailname);
 #endif
-	if((rawmailfd = open(mailfile, O_RDONLY)) < 0) {
+	if ((rawmailfd = open(mailfile, O_RDONLY)) < 0) {
 		unlink(donemailname);
 		free(donemailname);
 		log_error(LOG_ARGS, "could not open() input mail file");
@@ -484,7 +371,7 @@ int main(int argc, char **argv)
 	footfd = openat(ml.ctrlfd, "footer", O_RDONLY);
 
 	delheaders = ctrlvalues(ml.ctrlfd, "delheaders");
-	if(delheaders == NULL)
+	if (delheaders == NULL)
 		delheaders = xcalloc(1, sizeof(*delheaders));
 
 	tll_push_back(*delheaders, xstrdup("From "));
@@ -492,10 +379,14 @@ int main(int argc, char **argv)
 
 	subjectprefix = ctrlvalue(ml.ctrlfd, "prefix");
 
-	if(do_all_the_voodoo_here(rawmailfd, donemailfd, hdrfd, footfd,
+	replyto = statctrl(ml.ctrlfd, "replyto");
+
+	if (do_all_the_voodoo_here(rawmailfd, donemailfd, hdrfd, footfd,
 				delheaders, readhdrs,
-				&allheaders, subjectprefix) < 0) {
+				&allheaders, subjectprefix, replyto) < 0) {
 		log_error(LOG_ARGS, "Error in do_all_the_voodoo_here");
+		free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+			    toemails, ccemails, rpemails, dtemails, allheaders);
 		exit(EXIT_FAILURE);
 	}
 
@@ -504,9 +395,9 @@ int main(int argc, char **argv)
 	close(rawmailfd);
 	close(donemailfd);
 
-	if(hdrfd >= 0)
+	if (hdrfd >= 0)
 		close(hdrfd);
-	if(footfd >= 0)
+	if (footfd >= 0)
 		close(footfd);
 
 	/* To: addresses */
@@ -530,7 +421,7 @@ int main(int argc, char **argv)
 	}
 	addr_in_to_or_cc = !statctrl(ml.ctrlfd, "tocc");
 
-	if(addr_in_to_or_cc || findaddress) {
+	if (addr_in_to_or_cc || findaddress) {
 		listaddrs = ctrlvalues(ml.ctrlfd, "listaddress");
 		if (listaddrs == NULL) {
 			log_error(LOG_ARGS, "list address is not defined");
@@ -539,7 +430,7 @@ int main(int argc, char **argv)
 		findaddress = !find_email(&dtemails, listaddrs, ml.delim,
 		    recipextra != NULL ? NULL : &recipextra);
 	}
-	if(addr_in_to_or_cc || findaddress) {
+	if (addr_in_to_or_cc || findaddress) {
 		intocc = find_email(&toemails, listaddrs, ml.delim,
 		    recipextra != NULL ? NULL : &recipextra);
 		if (!intocc)
@@ -565,7 +456,7 @@ int main(int argc, char **argv)
 	}
 
 	/* discard malformed mail with invalid From: unless it's a bounce */
-	if(tll_length(fromemails) != 1 &&
+	if (tll_length(fromemails) != 1 &&
 			(recipextra == NULL ||
 			strncmp(recipextra, "bounces", 7) != 0)) {
 		tll_foreach(fromemails, it)
@@ -582,7 +473,8 @@ int main(int argc, char **argv)
 		free(donemailname);
 		free(discardname);
 		free(randomstr);
-		/* TODO: free emailstructs */
+		free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+			    toemails, ccemails, rpemails, dtemails, allheaders);
 		exit(EXIT_SUCCESS);
 	}
 	/* The only time posteraddr will remain unset is when the mail is a
@@ -597,10 +489,10 @@ int main(int argc, char **argv)
 	}
 
 	/* envelope from */
-	if((envstr = getenv("SENDER")) != NULL) {
+	if ((envstr = getenv("SENDER")) != NULL) {
 		/* qmail, postfix, exim */
 		efrom = envstr;
-	} else if(tll_length(rpemails) >= 1) {
+	} else if (tll_length(rpemails) >= 1) {
 		/* the (first) Return-Path: header */
 		efrom = tll_front(rpemails);
 	}
@@ -610,31 +502,37 @@ int main(int argc, char **argv)
 			subject = unistr_header_to_utf8(readhdrs[5].values[0]);
 	if (!subject) subject = xstrdup("");
 
-	if(recipextra) {
+	if (recipextra) {
 		int ownfd = openat(ml.ctrlfd, "owner", O_RDONLY);
 		xasprintf(&owner, "%d", ownfd);
-		if(strcmp(recipextra, "owner") == 0) {
+		if (strcmp(recipextra, "owner") == 0) {
 			/* Why is this here, and not in listcontrol() ?
 			 * -- mortenp 20060409 */
 			/* strip envelope from before resending */
 			tll_push_back(*delheaders, xstrdup("From "));
 			tll_push_back(*delheaders, xstrdup("Return-Path:"));
 
-			if((rawmailfd = open(mailfile, O_RDONLY)) < 0) {
+			if ((rawmailfd = open(mailfile, O_RDONLY)) < 0) {
 				log_error(LOG_ARGS, "could not open() "
 						    "input mail file");
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_FAILURE);
 			}
-			if((donemailfd = open(donemailname,
+			if ((donemailfd = open(donemailname,
 						O_WRONLY|O_TRUNC)) < 0) {
 				log_error(LOG_ARGS, "could not open() "
 						    "output mail file");
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_FAILURE);
 			}
-			if(do_all_the_voodoo_here(rawmailfd, donemailfd, -1,
+			if (do_all_the_voodoo_here(rawmailfd, donemailfd, -1,
 					-1, delheaders,
-					NULL, &allheaders, NULL) < 0) {
+					NULL, &allheaders, NULL, 0) < 0) {
 				log_error(LOG_ARGS, "do_all_the_voodoo_here");
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_FAILURE);
 			}
 			tll_free_and_free(*delheaders, free);
@@ -655,19 +553,23 @@ int main(int argc, char **argv)
 		listcontrol(testfrom, &ml, recipextra,
 			    mlmmjsend, donemailname);
 
+		free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+			    toemails, ccemails, rpemails, dtemails, allheaders);
 		return EXIT_SUCCESS;
 	}
 
 	/* checking incoming mail's size */
 	errno = 0;
 	maxmailsize = ctrlsizet(ml.ctrlfd, "maxmailsize", 0);
-	if(errno != ENOENT) {
-		if(stat(donemailname, &st) < 0) {
+	if (errno != ENOENT) {
+		if (stat(donemailname, &st) < 0) {
 			log_error(LOG_ARGS, "stat(%s,..) failed", donemailname);
+			free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+				    toemails, ccemails, rpemails, dtemails, allheaders);
 			exit(EXIT_FAILURE);
 		}
 
-		if(st.st_size > maxmailsize) {
+		if (st.st_size > maxmailsize) {
 			char *maxmailsizestr;
 			if (statctrl(ml.ctrlfd, "nomaxmailsizedenymails")) {
 				errno = 0;
@@ -677,6 +579,8 @@ int main(int argc, char **argv)
 				unlink(donemailname);
 				unlink(mailfile);
 				free(donemailname);
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_SUCCESS);
 			}
 
@@ -703,7 +607,7 @@ int main(int argc, char **argv)
 
 	free(delheaders);
 
-	if(*efrom == '\0') { /* don't send mails with <> in From
+	if (*efrom == '\0') { /* don't send mails with <> in From
 					     to the list */
 		xasprintf(&discardname, "%s/queue/discarded/%s", ml.dir, randomstr);
 		errno = 0;
@@ -714,13 +618,14 @@ int main(int argc, char **argv)
 		free(donemailname);
 		free(discardname);
 		free(randomstr);
-		/* TODO: free emailstructs */
+		free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+			    toemails, ccemails, rpemails, dtemails, allheaders);
 		exit(EXIT_FAILURE);
 	}
 
 	unlink(mailfile);
 
-	if(addr_in_to_or_cc && !intocc) {
+	if (addr_in_to_or_cc && !intocc) {
 		/* Don't send a mail about denial to the list, but silently
 		 * discard and exit. Also don't in case of it being turned off
 		 */
@@ -733,6 +638,8 @@ int main(int argc, char **argv)
 					mailfile);
 			unlink(donemailname);
 			free(donemailname);
+			free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+				    toemails, ccemails, rpemails, dtemails, allheaders);
 			exit(EXIT_SUCCESS);
 		}
 		txt = open_text(ml.fd, "deny", "post",
@@ -753,12 +660,18 @@ int main(int argc, char **argv)
 
 	access_rules = ctrlvalues(ml.ctrlfd, "access");
 	if (access_rules != NULL) {
-		enum action accret;
+		actions_t accret;
+		char *error = NULL;
+		char *qualifier = NULL;
 		/* Don't send a mail about denial to the list, but silently
 		 * discard and exit. Also do this in case it's turned off */
 		accret = do_access(access_rules, &allheaders,
-					posteraddr, ml.fd);
-		if (accret == DENY) {
+					posteraddr, &error, &qualifier);
+		if (error != NULL) {
+			log_oper(ml.fd, OPLOGFNAME, "%s", error);
+			free(error);
+		}
+		if (accret == ACT_DENY) {
 			if ((strcasecmp(ml.addr, posteraddr) == 0) ||
 				    statctrl(ml.ctrlfd, "noaccessdenymails")) {
 				log_error(LOG_ARGS, "Discarding %s because"
@@ -769,10 +682,12 @@ int main(int argc, char **argv)
 					mailfile);
 				unlink(donemailname);
 				free(donemailname);
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_SUCCESS);
 			}
 			txt = open_text(ml.fd, "deny", "post",
-					"access", NULL, "access");
+					"access", qualifier, "access");
 			MY_ASSERT(txt);
 			register_default_unformatted(txt, &ml);
 			register_unformatted(txt, "subject", subject);
@@ -786,28 +701,33 @@ int main(int argc, char **argv)
 			free(donemailname);
 			free(randomstr);
 			send_help(&ml, queuefilename, posteraddr);
-		} else if (accret == MODERATE) {
+		} else if (accret == ACT_MODERATE) {
 			moderated = 1;
 			modreason = ACCESS;
-		} else if (accret == DISCARD) {
+		} else if (accret == ACT_DISCARD) {
 			xasprintf(&discardname, "%s/queue/discarded/%s", ml.dir,
 			    randomstr);
 			free(randomstr);
-                	if(rename(donemailname, discardname) < 0) {
+                	if (rename(donemailname, discardname) < 0) {
 				log_error(LOG_ARGS, "could not rename(%s,%s)",
 					    donemailname, discardname);
 				free(donemailname);
 				free(discardname);
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_FAILURE);
 			}
 			free(donemailname);
 			free(discardname);
+			free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+				    toemails, ccemails, rpemails, dtemails, allheaders);
                 	exit(EXIT_SUCCESS);
-		} else if (accret == SEND) {
+		} else if (accret == ACT_SEND) {
 			send = 1;
-		} else if (accret == ALLOW) {
+		} else if (accret == ACT_ALLOW) {
 			/* continue processing as normal */
 		}
+		free(qualifier);
 	}
 
 	list_rules = ctrlvalues(ml.ctrlfd, "send");
@@ -826,7 +746,7 @@ int main(int argc, char **argv)
 	/* modnonsubposts implies subonlypost if modonlypost is not set */
 	if (modnonsubposts && !modonlypost) subonlypost = 1;
 
-	if(!send && (subonlypost || modonlypost || modnonsubposts)) {
+	if (!send && (subonlypost || modonlypost || modnonsubposts)) {
 		/* Don't send a mail about denial to the list, but silently
 		 * discard and exit. */
 		char *testaddr = posteraddr;
@@ -839,23 +759,25 @@ int main(int argc, char **argv)
 					mailfile);
 			unlink(donemailname);
 			free(donemailname);
+			free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+				    toemails, ccemails, rpemails, dtemails, allheaders);
 			exit(EXIT_SUCCESS);
 		}
-		if(subonlypost) {
+		if (subonlypost) {
 			foundaddr = (is_subbed(ml.fd, testaddr, 0) !=
 					SUB_NONE);
 		} else if (modonlypost) {
 			foundaddr = is_moderator(ml.fd, testaddr, NULL);
 		}
-		if(!foundaddr) {
-			if(modnonsubposts) {
+		if (!foundaddr) {
+			if (modnonsubposts) {
 			    moderated = 1;
 			    if (subonlypost)
 				modreason = MODNONSUBPOSTS;
 			    else if (modonlypost)
 				modreason = MODNONMODPOSTS;
 			} else {
-			    if((subonlypost &&
+			    if ((subonlypost &&
 				    statctrl(ml.ctrlfd, "nosubonlydenymails")) ||
 				    (modonlypost &&
 				    statctrl(ml.ctrlfd, "nomodonlydenymails"))) {
@@ -864,6 +786,8 @@ int main(int argc, char **argv)
 					mailfile);
 				unlink(donemailname);
 				free(donemailname);
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_SUCCESS);
 			    }
 			    if (subonlypost) {
@@ -889,8 +813,8 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if(!send && !moderated) {
-		if(statctrl(ml.ctrlfd, "moderated")) {
+	if (!send && !moderated) {
+		if (statctrl(ml.ctrlfd, "moderated")) {
 			moderated = 1;
 			modreason = MODERATED;
 		}
@@ -898,14 +822,16 @@ int main(int argc, char **argv)
 
 	notmetoo = statctrl(ml.ctrlfd, "notmetoo");
 
-	if(moderated) {
+	if (moderated) {
 		xasprintf(&mqueuename, "%s/moderation/%s", ml.dir, randomstr);
 		free(randomstr);
-		if(rename(donemailname, mqueuename) < 0) {
+		if (rename(donemailname, mqueuename) < 0) {
 			log_error(LOG_ARGS, "could not rename(%s,%s)",
 					    donemailname, mqueuename);
 			free(donemailname);
 			free(mqueuename);
+			free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+				    toemails, ccemails, rpemails, dtemails, allheaders);
 			exit(EXIT_FAILURE);
 		}
 		free(donemailname);
@@ -917,13 +843,17 @@ int main(int argc, char **argv)
 					    	    omitfilename);
 				free(mqueuename);
 				free(omitfilename);
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_FAILURE);
 			}
 			free(omitfilename);
-			if(dprintf(omitfd, "%s", posteraddr) < 0) {
+			if (dprintf(omitfd, "%s", posteraddr) < 0) {
 				log_error(LOG_ARGS,
 						"could not write omit file");
 				free(mqueuename);
+				free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+					    toemails, ccemails, rpemails, dtemails, allheaders);
 				exit(EXIT_FAILURE);
 			}
 			fsync(omitfd);
@@ -931,14 +861,17 @@ int main(int argc, char **argv)
 		}
 		newmoderated(&ml, mqueuename, mlmmjsend, efrom, subject,
 		    posteraddr, modreason);
+		free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+			    toemails, ccemails, rpemails, dtemails, allheaders);
 		return EXIT_SUCCESS;
 	}
 
 	free(randomstr);
 
-	if(noprocess) {
+	if (noprocess) {
 		free(donemailname);
-		/* XXX: toemails and ccemails etc. have to be free() */
+		free_parsed_hdrs(readhdrs, fromemails, originalfromemails,
+			    toemails, ccemails, rpemails, dtemails, allheaders);
 		exit(EXIT_SUCCESS);
 	}
 
