@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2021-2024 Baptiste Daroussin <bapt@FreeBSD.org>
+ * Copyright (C) 2021-2025 Baptiste Daroussin <bapt@FreeBSD.org>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -84,9 +84,9 @@ send_probe(struct ml *ml, const char *addr)
 	txt = open_text(ml->fd, "probe", NULL, NULL, NULL, "bounce-probe");
 	MY_ASSERT(txt);
 	register_default_unformatted(txt, ml);
-	fls = init_file_lines_fd(openat(bfd, addr, O_RDONLY), ':');
+	fls = init_file_lines_data(openat(bfd, addr, O_RDONLY), ':', ml);
 	register_formatted(txt, "bouncenumbers",
-			rewind_file_lines, get_file_line, fls);
+			rewind_file_lines, get_msgid_line, fls);
 	queuefilename = prepstdreply(txt, ml, "$listowner$", myaddr, NULL);
 	MY_ASSERT(queuefilename);
 	close_text(txt);
@@ -272,7 +272,7 @@ open_subscriber_directory(int listfd, enum subtype typesub, const char **subdir)
 		*subdir = dir;
 	if (dir == NULL)
 		return (-1);
-	return (openat(listfd, dir, O_DIRECTORY|O_CLOEXEC));
+	return (openat(listfd, dir, O_DIRECTORY|O_CLOEXEC|O_RDONLY));
 }
 
 bool
@@ -399,7 +399,7 @@ unsubscribe(int listfd, const char *address, enum subtype typesub)
 		xasprintf(&wrname, "%s.new", dp->d_name);
 		wfd = openat(fd, wrname, O_RDWR|O_CREAT|O_TRUNC,
 		    S_IRUSR|S_IWUSR|groupwritable);
-		if (wfd == -1 && !lock(wfd, true)) {
+		if (wfd == -1 || !lock(wfd, true)) {
 			log_err("Could not open '%s/%s': %s", subdir, wrname,
 			    strerror(errno));
 			munmap(start, st.st_size);
@@ -710,6 +710,34 @@ ml_open(struct ml *ml, bool checkuser)
 	return (true);
 }
 
+void
+save_hdr_if_needed(const char *line, struct mailhdr *readhdrs)
+{
+	for (size_t i = 0; readhdrs[i].token != NULL; i++) {
+		size_t tokenlen = strlen(readhdrs[i].token);
+		if (strncasecmp(line, readhdrs[i].token, tokenlen) == 0) {
+			readhdrs[i].valuecount++;
+			readhdrs[i].values =
+				(char **)xrealloc(readhdrs[i].values,
+				  readhdrs[i].valuecount * sizeof(char *));
+			readhdrs[i].values[readhdrs[i].valuecount - 1] =
+				xstrdup(line + tokenlen);
+		}
+	}
+}
+
+void
+free_mailhdrs(struct mailhdr *hdrs)
+{
+	for (int i = 0; hdrs[i].token != NULL; i++) {
+		for (int j = 0; j < hdrs[i].valuecount; j++)
+			free(hdrs[i].values[j]);
+		free(hdrs[i].values);
+		hdrs[i].values = NULL;
+		hdrs[i].valuecount = 0;
+	}
+}
+
 void
 ml_close(struct ml *ml)
 {
