@@ -45,6 +45,9 @@
 #include "xstring.h"
 #include "ctrlvalues.h"
 
+#include "find_email_adr.h"
+#include "do_all_the_voodoo_here.h"
+
 char *subtype_strs[] = {
 	"normal",
 	"digest",
@@ -92,7 +95,7 @@ find_subscriber(int fd, const char *address)
 	}
 	free(line);
 	fclose(f);
-	return (ret);;
+	return (ret);
 }
 
 int
@@ -136,19 +139,19 @@ is_subbed(int listfd, const char *address, bool both)
 	enum subtype typesub = SUB_NONE;
 	int fd;
 
-	fd = openat(listfd, "subscribers.d", O_DIRECTORY);
+	fd = openat(listfd, "subscribers.d", O_DIRECTORY|O_RDONLY);
 	if (fd != -1 && is_subbed_in(fd, "subscribers.d", address)) {
 		if (!both) return SUB_NORMAL;
 		typesub = SUB_NORMAL;
 	}
 
-	fd = openat(listfd, "digesters.d", O_DIRECTORY);
+	fd = openat(listfd, "digesters.d", O_DIRECTORY|O_RDONLY);
 	if (fd != -1 && is_subbed_in(fd, "digesters.d", address)) {
 		if (typesub == SUB_NORMAL) return SUB_BOTH;
 		return SUB_DIGEST;
 	}
 
-	fd = openat(listfd, "nomailsubs.d", O_DIRECTORY);
+	fd = openat(listfd, "nomailsubs.d", O_DIRECTORY|O_RDONLY);
 	if (fd != -1 && is_subbed_in(fd, "nomailsubs.d", address))
 		return SUB_NOMAIL;
 
@@ -631,7 +634,7 @@ subscribe_type(int listfd, char *address, enum subtype typesub)
 
 	subfilefd = openat(dirfd, chstr, O_RDWR|O_CREAT|O_APPEND,
 				S_IRUSR|S_IWUSR|groupwritable);
-	if(subfilefd == -1 && !lock(subfilefd, true)) {
+	if(subfilefd == -1 || !lock(subfilefd, true)) {
 		log_error(LOG_ARGS, "Could not open '%s/%s'", subdir, chstr);
 		exit(EXIT_FAILURE);
 	}
@@ -720,3 +723,36 @@ do_subscribe(struct ml *ml, struct subscription *sub, const char *addr)
 	free(address);
 	return (true);
 }
+
+void
+autosubscribe_sender(struct ml *ml, int mailfd, enum subtype typesub)
+{
+	struct mailhdr readhdrs[] = {
+		{ "From:", 0, NULL },
+		{ NULL, 0, NULL }
+	};
+	strlist from = tll_init();
+
+	FILE *f = fdopen(mailfd, "r");
+	scan_headers(f, readhdrs, NULL, NULL);
+	fclose(f);
+	if (readhdrs[0].valuecount == 0) {
+		free_mailhdrs(readhdrs);
+		return;
+	}
+	find_email_adr(readhdrs[0].values[0], &from);
+	if (tll_length(from) == 0) {
+		tll_free(from);
+		free_mailhdrs(readhdrs);
+		return;
+	}
+	if (strchr(tll_front(from), '@') == NULL) {
+		tll_free_and_free(from, free);
+		free_mailhdrs(readhdrs);
+		return;
+	}
+	subscribe_type(ml->fd, tll_front(from), typesub);
+	tll_free_and_free(from, free);
+	free_mailhdrs(readhdrs);
+}
+
